(* module Primitives *)

open Lolevel.Error
open Types
open Hilevel.Printing

let get_some = function
  | Some (_,x) -> x
  | None -> invalid_arg "get_some"

let always = fun res -> Some res

(*
(* lists not implemented *)
let rec _Nil : constructor =
  let printer _ = print "[]" in
  let projector = function
    | IsNil -> Some(void_fi,ResLiteral(Bool(true)))
    | Tail -> Some(void_fi,ResConstructor _Nil)
    | _ -> None in
  Cons0(printer, projector)

let _Cons : constructor =
  let printer pr1 pr2 _ = () (* FIXME *) in
  let projector arg1 arg2 = function
    | IsZero -> Some(void_fi,ResLiteral(Bool(false)))
    | Head -> Some(arg1)
    | Tail -> Some(arg2)
    | _ -> None in
  let check1 = always in
  let check2 _arg1 = function
    | _,ResConstructor(Cons0(_, proj)) as arg2 ->
        (match proj IsNil with Some _ -> Some arg2 | _ -> None)
    | _ -> None in
  Cons2(printer, projector, check1, check2)
*)

let rec _Zero : constructor =
  let printer _ = print "0." in
  let projector = function
    | IsZero -> Some(void_fi,ResLiteral(Bool(true)))
    | Pred -> Some(void_fi,ResConstructor(_Zero))
    | _ -> None in
  Cons0(printer, projector)

let _Succ : constructor =
  let printer _ = function
    | None -> print "succ"
    | Some proj ->
        let rec aux n (ResConstructor(Cons0(_, proj))) =
          match proj IsZero with
          | Some(_,ResLiteral(Bool(true))) -> print_int n; print "."
          | _ -> aux (n+1) (get_some (proj Pred)) in
      aux 1 (get_some (proj Pred)) in
  let projector arg1 = function
    | IsZero -> Some(void_fi,ResLiteral(Bool(false)))
    | Pred -> Some(arg1) (* use predecessor's fi *)
    | _ -> None in
  let check1 = function
    | _,ResConstructor(Cons0(_, proj)) as arg1 ->
        (match proj IsZero with Some _ -> Some arg1 | _ -> None)
    | _ -> None in
  Cons1(printer, projector, check1)

let _Unit : constructor =
  let printer _ = print "()" in
  let projector = function _ -> None in
  Cons0(printer, projector)

let _Box : constructor =
  let printer pr1 = function
    | None -> print "box"
    | Some _ -> begin
        obox();
        print "(box ";
        pr1();
        print ")";
        cbox()
      end in
  let projector arg1 = function
    | Fst -> Some arg1
    | _ -> None in
  let check1 = always in
  Cons1(printer, projector, check1)

let _Pair : constructor =
  let printer pr1 pr2 _ = begin
      obox();
      print "(";
      pr1();
      print ",";
      break();
      pr2();
      print ")";
      cbox()
    end in
  let projector arg1 arg2 = function
    | Fst -> Some arg1
    | Snd -> Some arg2
    | _ -> None in
  let check1 = always in
  let check2 _arg1 = always in
  Cons2(printer, projector, check1, check2)

let _Triple : constructor =
  let printer pr1 pr2 pr3 _ = begin
      obox();
      print "(";
      pr1();
      print ",";
      break();
      pr2();
      print ",";
      break();
      pr3();
      print ")";
      cbox()
    end in
  let projector arg1 arg2 arg3 = function
    | Fst -> Some arg1
    | Snd -> Some arg2
    | Trd -> Some arg3
    | _ -> None in
  let check1 = always in
  let check2 _arg1 = always in
  let check3 _arg1 _arg2 = always in
  Cons3(printer, projector, check1, check2, check3)

let _Not : primfun =
  let printer () = print "not" in
  let gobble1 = function
    | _,ResLiteral(Bool true) -> Some(void_fi,ResLiteral(Bool(false)))
    | _,ResLiteral(Bool false) -> Some(void_fi,ResLiteral(Bool(true)))
    | _ -> None in
  Funct1(printer, gobble1)

let _Leq : primfun =
  let printer pr1 () = begin
      obox();
      print "(";
      pr1();
      print "<=)";
      cbox()
    end in
  let gobble1 = function
    | _,ResLiteral(Num n) as arg -> Some(arg)
    | _ -> None in
  let gobble2 (_,ResLiteral(Num arg1)) = function
    | _,ResLiteral(Num n) -> Some(void_fi,ResLiteral(Bool(arg1<=n)))
    | _ -> None in
  Funct2(printer, gobble1, gobble2)

let _Lt : primfun =
  let printer pr1 () = begin
      obox();
      print "(";
      pr1();
      print "<)";
      cbox()
    end in
  let gobble1 = function
    | _,ResLiteral(Num n) as arg -> Some(arg)
    | _ -> None in
  let gobble2 (_,ResLiteral(Num arg1)) = function
    | _,ResLiteral(Num n) -> Some(void_fi,ResLiteral(Bool(arg1<n)))
    | _ -> None in
  Funct2(printer, gobble1, gobble2)

let _Eq : primfun =
  let printer pr1 () = begin
      obox();
      print "(";
      pr1();
      print "==)";
      cbox()
    end in
  let gobble1 = function
    | _,ResLiteral(Num n) as arg -> Some(arg)
    | _ -> None in
  let gobble2 (_,ResLiteral(Num arg1)) = function
    | _,ResLiteral(Num n) -> Some(void_fi,ResLiteral(Bool(arg1==n)))
    | _ -> None in
  Funct2(printer, gobble1, gobble2)

let _Plus : primfun =
  let printer pr1 () = begin
      obox();
      print "(";
      pr1();
      print "+)";
      cbox()
    end in
  let gobble1 = function
    | _,ResLiteral(Num n) as arg -> Some(arg)
    | _ -> None in
  let gobble2 (_,ResLiteral(Num arg1)) = function
    | _,ResLiteral(Num n) -> Some(void_fi,ResLiteral(Num(arg1+n)))
    | _ -> None in
  Funct2(printer, gobble1, gobble2)

let _Minus : primfun =
  let printer pr1 () = begin
      obox();
      print "(";
      pr1();
      print "-)";
      cbox()
    end in
  let gobble1 = function
    | _,ResLiteral(Num n) as arg -> Some(arg)
    | _ -> None in
  let gobble2 (_,ResLiteral(Num arg1)) = function
    | _,ResLiteral(Num n) -> Some(void_fi,ResLiteral(Num(if arg1 < n then 0 else arg1-n)))
    | _ -> None in
  Funct2(printer, gobble1, gobble2)

let _Times : primfun =
  let printer pr1 () = begin
      obox();
      print "(";
      pr1();
      print "*)";
      cbox()
    end in
  let gobble1 = function
    | _,ResLiteral(Num n) as arg -> Some(arg)
    | _ -> None in
  let gobble2 (_,ResLiteral(Num arg1)) = function
    | _,ResLiteral(Num n) -> Some(void_fi,ResLiteral(Num(arg1*n)))
    | _ -> None in
  Funct2(printer, gobble1, gobble2)
